import { Types } from "mongoose";

import { AccountModel } from "@/models/account.model";
import { RoleModel } from "@/models/role.model";
import { PermissionModule } from "@/services/permission/permission.module";

import { getUser } from "@/clients/auth";
import { validateEmptyFields } from "@/utils/validate-empty-fields";
import { determinePaginationParameters } from "@/utils/determine-pagination-parameters";

import type { FastifyReply, FastifyRequest } from "fastify";
import type { Account } from "@/types/models/account.type";
import type { GetAccountRequest, SetAccountRoleRequest, GetAccountsRequest, GetAccountsResponse } from "./types";

export const getAccounts = async (request: GetAccountsRequest): Promise<GetAccountsResponse> => {
  const { page, limit } = determinePaginationParameters(request?.query ?? {});

  const accountsCount = await AccountModel.countDocuments();

  const totalPages = Math.ceil(accountsCount / limit);
  const offset = (page - 1) * limit;

  const accounts = await AccountModel.find({}).sort({ createdAt: "desc" }).skip(offset).limit(limit).lean();

  return { accounts, totalPages };
};

export const createAccount = async (request: FastifyRequest, reply: FastifyReply) => {
  if (!Types.ObjectId.isValid(request.user.id)) {
    void reply.status(400);
    return new Error("invalid user id");
  }

  const account = await AccountModel.findOne({ userId: request.user.id }).lean();

  if (account) {
    void reply.status(409);
    return new Error("account already exist");
  }

  const user = await getUser({ id: request.user.id });

  if (!user) {
    void reply.status(404);
    return new Error("user not found");
  }

  const createdAccount = new AccountModel({
    userId: user._id,
  });

  return createdAccount.save();
};

export const getAccountByID = async (request: GetAccountRequest, reply: FastifyReply): Promise<Account> => {
  const [getAccountRequestParams, error] = validateEmptyFields(request.params ?? {}, ["userId"]);

  if (error) {
    void reply.status(400);
    throw error;
  }

  if (!Types.ObjectId.isValid(getAccountRequestParams.userId)) {
    void reply.status(400);
    throw new Error("invalid user id");
  }

  const account = await AccountModel.findOne({ userId: getAccountRequestParams.userId }).lean();

  if (!account) {
    void reply.status(404);
    throw new Error("account not found");
  }

  return account;
};

export const getAccount = async (request: GetAccountRequest, reply: FastifyReply): Promise<Account> => {
  if (!Types.ObjectId.isValid(request.user.id)) {
    void reply.status(400);
    throw new Error("invalid user id");
  }

  const account = await AccountModel.findOne({ userId: request.user.id }).lean();

  if (!account) {
    void reply.status(404);
    throw new Error("account not found");
  }

  return account;
};

export const setAccountRole = async (request: SetAccountRoleRequest, reply: FastifyReply) => {
  const [setAccountRoleBody, error] = validateEmptyFields(request.body || {}, ["userId", "role"]);

  if (error) {
    void reply.status(400);
    return error;
  }

  const role = await RoleModel.findOne({ name: setAccountRoleBody.role }).lean();

  if (!role) {
    void reply.status(404);
    return new Error("role not found");
  }

  const account = await AccountModel.findOneAndUpdate(
    { userId: setAccountRoleBody.userId },
    { $set: { role: role.name } }
  );

  if (!account) {
    void reply.status(404);
    return new Error("account not found");
  }

  return account;
};

export const removeAccount = async (request: FastifyRequest, reply: FastifyReply): Promise<Account> => {
  if (!Types.ObjectId.isValid(request.user.id)) {
    void reply.status(400);
    throw new Error("invalid user id");
  }

  const account = await AccountModel.findOneAndUpdate(
    { userId: request.user.id },
    { $set: { isDeleted: true, deletedAt: new Date() } }
  ).lean();

  if (!account) {
    void reply.status(404);
    throw new Error("account not found");
  }

  return account;
};

export const removeAccountById = async (request: GetAccountRequest, reply: FastifyReply): Promise<Account> => {
  const [{ userId }, error] = validateEmptyFields(request.params ?? {}, ["userId"]);

  if (error) {
    void reply.status(400);
    throw error;
  }

  if (!Types.ObjectId.isValid(userId)) {
    void reply.status(400);
    throw new Error("invalid user id");
  }

  const account = await AccountModel.findOne({ userId: request.user.id });

  if (!account) {
    void reply.status(404);
    throw new Error("account not found");
  }

  const isAvailableToDelete = await PermissionModule.deleteAccount(account);

  if (!isAvailableToDelete) {
    throw new Error("the user doesn't have the permission to delete account");
  }

  await account.updateOne({ $set: { isDeleted: true, deletedAt: new Date() } });

  return account;
};

export const deleteAccount = async (request: FastifyRequest, reply: FastifyReply): Promise<Account> => {
  if (!Types.ObjectId.isValid(request.user.id)) {
    void reply.status(400);
    throw new Error("invalid user id");
  }

  const account = await AccountModel.findByIdAndDelete({ userId: request.user.id }).lean();

  if (!account) {
    void reply.status(404);
    throw new Error("account not found");
  }

  return account;
};

export const deleteAccountById = async (request: GetAccountRequest, reply: FastifyReply): Promise<Account> => {
  const [{ userId }, error] = validateEmptyFields(request.params ?? {}, ["userId"]);

  if (error) {
    void reply.status(400);
    throw error;
  }

  if (!Types.ObjectId.isValid(userId)) {
    void reply.status(400);
    throw new Error("invalid user id");
  }

  const account = await AccountModel.findByIdAndDelete({ userId: request.user.id }).lean();

  if (!account) {
    void reply.status(404);
    throw new Error("account not found");
  }

  const isAvailableToDelete = await PermissionModule.deleteAccount(account);

  if (!isAvailableToDelete) {
    throw new Error("the user doesn't have the permission to delete account");
  }

  return account;
};

export const restoreAccount = async (request: FastifyRequest, reply: FastifyReply): Promise<Account> => {
  if (!Types.ObjectId.isValid(request.user.id)) {
    void reply.status(400);
    throw new Error("invalid user id");
  }

  const account = await AccountModel.findOneAndUpdate(
    { userId: request.user.id },
    { $set: { isDeleted: false } }
  ).lean();

  if (!account) {
    void reply.status(404);
    throw new Error("account not found");
  }

  return account;
};
