import { Types } from "mongoose";

import { PermissionModule } from "@/services/permission/permission.module";
import { PermissionModel } from "@/models/permission.model";

import { validateEmptyFields } from "@/utils/validate-empty-fields";

import type { FastifyReply } from "fastify";
import type { Permission } from "@/types/models/permission.type";
import type { CreatePermissionRequest, GetPermissionByIdRequest, UpdatePermissionRequest } from "./types";

export const getAllPermissions = async (): Promise<Permission[]> => PermissionModule.getAllPermissions();

export const getPermissionById = async (
  request: GetPermissionByIdRequest,
  reply: FastifyReply
): Promise<Permission | undefined> => {
  const [{ permissionId }, validateParamsError] = validateEmptyFields(request.params ?? {}, ["permissionId"]);

  if (validateParamsError) {
    void reply.status(400);
    throw validateParamsError;
  }

  if (!Types.ObjectId.isValid(permissionId)) {
    void reply.status(400);
    throw new Error("invalid permission id");
  }

  const permission = await PermissionModel.findOne({
    _id: new Types.ObjectId(permissionId),
    isDeleted: false,
  });

  if (!permission) {
    void reply.status(404);
    throw new Error("permission not found");
  }

  return permission;
};

export const createPermission = async (request: CreatePermissionRequest, reply: FastifyReply): Promise<Permission> => {
  const [permission, error] = validateEmptyFields(request.body ?? {}, ["name", "description"]);

  if (error) {
    void reply.status(400);
    throw error;
  }

  const newPermission = new PermissionModel(permission);

  return newPermission.save();
};

export const updatePermission = async (
  request: UpdatePermissionRequest,
  reply: FastifyReply
): Promise<Permission | undefined> => {
  const [permission, validateBodyError] = validateEmptyFields(request.body ?? {}, ["name", "description"]);
  const [{ permissionId }, validateParamsError] = validateEmptyFields(request.params ?? {}, ["permissionId"]);

  if (validateBodyError) {
    void reply.status(400);
    throw validateBodyError;
  }

  if (validateParamsError) {
    void reply.status(400);
    throw validateParamsError;
  }

  if (!Types.ObjectId.isValid(permissionId)) {
    void reply.status(400);
    throw new Error("invalid permission id");
  }

  const findedPermission = await PermissionModel.findByIdAndUpdate(permissionId, {
    $set: {
      ...permission,
      updatedAt: new Date(),
    },
  });

  if (!findedPermission) {
    void reply.status(404);
    throw new Error("permission not found");
  }

  return permission;
};

export const removePermission = async (
  request: GetPermissionByIdRequest,
  reply: FastifyReply
): Promise<Permission | undefined> => {
  const [{ permissionId }, validateParamsError] = validateEmptyFields(request.params ?? {}, ["permissionId"]);

  if (validateParamsError) {
    void reply.status(400);
    throw validateParamsError;
  }

  if (!Types.ObjectId.isValid(permissionId)) {
    void reply.status(400);
    throw new Error("invalid permission id");
  }

  const permission = await PermissionModel.findOneAndUpdate(
    {
      _id: new Types.ObjectId(permissionId),
      isDeleted: false,
    },
    {
      $set: {
        deletedAt: new Date(),
        isDeleted: true,
      },
    }
  );

  if (!permission) {
    void reply.status(404);
    throw new Error("permission not found");
  }

  return permission;
};

export const restorePermission = async (
  request: GetPermissionByIdRequest,
  reply: FastifyReply
): Promise<Permission | undefined> => {
  const [{ permissionId }, validateParamsError] = validateEmptyFields(request.params ?? {}, ["permissionId"]);

  if (validateParamsError) {
    void reply.status(400);
    throw validateParamsError;
  }

  if (!Types.ObjectId.isValid(permissionId)) {
    void reply.status(400);
    throw new Error("invalid permission id");
  }

  const permission = await PermissionModel.findOneAndUpdate(
    {
      _id: new Types.ObjectId(permissionId),
      isDeleted: true,
    },
    {
      $set: {
        deletedAt: undefined,
        updatedAt: new Date(),
        isDeleted: false,
      },
    }
  );

  if (!permission) {
    void reply.status(404);
    throw new Error("permission not found");
  }

  return permission;
};
