import { Types } from "mongoose";

import { RoleModel } from "@/models/role.model";

import { convertArrayToRecord } from "@/utils/convert-array-to-record";
import { validateEmptyFields } from "@/utils/validate-empty-fields";

import type { FastifyReply } from "fastify";
import type { CreateRoleRequest, GetRoleRequest, RemoveRoleRequest, UpdateRoleRequest } from "./types";

export const getAllRoles = async () => RoleModel.find({ isDeleted: false }).lean();

export const createRole = async (request: CreateRoleRequest, reply: FastifyReply) => {
  const [createRoleRequest, error] = validateEmptyFields(request.body, ["name", "permissions"]);

  if (error || !createRoleRequest) {
    void reply.status(400);
    return error;
  }

  const isRoleExist = await RoleModel.exists({ name: createRoleRequest.name }).lean();

  if (isRoleExist) {
    void reply.status(409);
    return new Error("role already exist");
  }

  const permissionsRecord = convertArrayToRecord(createRoleRequest.permissions, true);

  return RoleModel.create({ name: createRoleRequest.name, permissions: permissionsRecord });
};

export const deleteRole = async (request: RemoveRoleRequest, reply: FastifyReply) => {
  const { id } = request.body || {};

  if (!id || !Types.ObjectId.isValid(id)) {
    void reply.status(400);
    return new Error("wrong id");
  }

  const deletedRole = await RoleModel.findByIdAndDelete(id);

  if (!deletedRole) {
    void reply.status(404);
    return new Error("role not found");
  }

  return deletedRole.toJSON();
};

export const removeRole = async (request: RemoveRoleRequest, reply: FastifyReply) => {
  const { id } = request.body || {};

  if (!id || !Types.ObjectId.isValid(id)) {
    void reply.status(400);
    return new Error("wrong id");
  }

  const role = await RoleModel.findById(id);

  if (!role) {
    void reply.status(404);
    return new Error("role by id not found");
  }

  if (role.isDeleted) {
    void reply.status(409);
    return new Error("role already deleted");
  }

  await role.updateOne({ $set: { isDeleted: true, deletedAt: Date.now() } });

  return role;
};

export const getRole = async (request: GetRoleRequest, reply: FastifyReply) => {
  const { query } = request.params || {};

  if (!query) {
    void reply.status(400);
    return new Error("query is empty");
  }

  if (Types.ObjectId.isValid(query)) {
    const role = await RoleModel.findOne({ _id: new Types.ObjectId(query), isDeleted: false }).lean();

    if (!role) {
      void reply.status(404);
      return new Error("role not found");
    }

    return role;
  }

  const role = await RoleModel.findOne({ name: query, isDeleted: false }).lean();

  if (!role) {
    void reply.status(404);
    return new Error("role not found");
  }

  return role;
};

export const replaceRole = async (request: UpdateRoleRequest, reply: FastifyReply) => {
  const [replaceRoleRequest, error] = validateEmptyFields(request.body ?? {}, ["name", "permissions"]);
  const { query } = request.params || {};

  if (error || !replaceRoleRequest) {
    void reply.status(400);
    return error;
  }

  if (!query) {
    void reply.status(400);
    return new Error("either name or id must be filled");
  }

  if (Types.ObjectId.isValid(query)) {
    const role = await RoleModel.findById(query);

    if (!role) {
      void reply.status(404);
      return new Error("role by id not found");
    }

    await role.replaceOne({
      name: replaceRoleRequest.name,
      permissions: convertArrayToRecord(replaceRoleRequest.permissions, true),
      updatedAt: Date.now(),
    });

    return role;
  }

  const role = await RoleModel.findOne({ name: query });

  if (!role) {
    void reply.status(404);
    return new Error("role not found");
  }

  await role.replaceOne({
    name: replaceRoleRequest.name,
    permissions: convertArrayToRecord(replaceRoleRequest.permissions, true),
    updatedAt: Date.now(),
  });

  return role;
};

export const updateRole = async (request: UpdateRoleRequest, reply: FastifyReply) => {
  const { query } = request.params;

  if (!query) {
    void reply.status(400);
    return new Error("query is empty");
  }

  if (!request.body?.name && !request.body?.permissions) {
    void reply.status(400);
    return new Error("either name or permissions must be filled");
  }

  if (Types.ObjectId.isValid(query)) {
    const role = await RoleModel.findById(query);

    if (!role) {
      void reply.status(404);
      return new Error("role by id not found");
    }

    await role.updateOne({
      $set: {
        name: request.body?.name,
        permissions: request.body?.permissions && convertArrayToRecord(request.body.permissions, true),
        updatedAt: Date.now(),
      },
    });

    return role;
  }

  const role = await RoleModel.findOne({ name: query });

  if (!role) {
    void reply.status(404);
    return new Error("role not found");
  }

  await role.updateOne({
    $set: {
      name: request.body?.name,
      permissions: request.body?.permissions && convertArrayToRecord(request.body.permissions, true),
      updatedAt: Date.now(),
    },
  });

  return role;
};

export const restoreRole = async (request: RemoveRoleRequest, reply: FastifyReply) => {
  const { id } = request.body || {};

  if (!id || !Types.ObjectId.isValid(id)) {
    void reply.status(400);
    return new Error("wrong id");
  }

  const role = await RoleModel.findById(id);

  if (!role) {
    void reply.status(404);
    return new Error("role by id not found");
  }

  if (!role.isDeleted) {
    void reply.status(409);
    return new Error("role not removed");
  }

  await role.updateOne({ $set: { isDeleted: false } });

  return role;
};
