import fastify, { FastifyInstance } from "fastify";
import { connect as mongoConnect } from "mongoose";

import { Router } from "./router";

import { registerFastifyPlugins } from "@/configuration/register-fastify-plugins";
import { combineRoutes } from "@/configuration/combine-routes";
import { constituteMongoURI } from "@/configuration/constitute-mongo-uri";

import { DEFAULT } from "@/constants/default";

import type { PluginsOptions } from "@/types/configuration/plugins-options";
import type { DatabaseOptions } from "@/types/configuration/database-options";

type ServerOptions = {
  port?: number;
  host?: string;
  backlog?: number;
};

type ServerArgs = {
  serverOptions?: ServerOptions;
  pluginsOptions?: PluginsOptions;
  databaseOptions?: DatabaseOptions;
};

export class Server {
  private fastify: FastifyInstance;
  private serverOptions?: ServerOptions;
  private databaseOptions?: DatabaseOptions;

  /**
   * Настройка сервера
   * @param {ServerOptions} serverOptions объект настроек сервера.
   * @param {DatabaseOptions} databaseOptions объект настроек подключения к базе данных.
   * @param {PluginsOptions} pluginsOptions объект настроек плагинов fastify.
   */
  constructor({ serverOptions, databaseOptions, pluginsOptions }: ServerArgs) {
    this.serverOptions = serverOptions;
    this.databaseOptions = databaseOptions;
    this.fastify = fastify(DEFAULT.fastifyOptions);

    if (pluginsOptions) {
      registerFastifyPlugins(this.fastify, pluginsOptions);
    }

    combineRoutes(new Router(this.fastify));
  }

  public start = async () => {
    const fasticyConnection = this.fastify.listen(this.serverOptions);
    const databaseConnection = this.databaseOptions ? mongoConnect(constituteMongoURI(this.databaseOptions)) : null;

    await Promise.all([databaseConnection, fasticyConnection])
      .then(() => {
        this.fastify.swagger();
        console.info(`server started on ${this.serverOptions?.host}:${this.serverOptions?.port}`);
      })
      .catch((reason) => console.error(reason));
  };
}
