import { mock, mockReset } from "jest-mock-extended";

import { PermissionService } from "./permission.service";

import { PERMISSIONS } from "@/constants/permissions";

import type { AccountService } from "./permission.interface";
import type { Account } from "@/types/models/account.type";

const DEFAULT_ACCOUNT: Account = {
  userId: "userId",
  nickname: "nickname",
  avatar: "/img/avatar.png",
  role: "admin",
  createdAt: new Date(),
  updatedAt: new Date(),
  isDeleted: false,
};

describe("PermissionService", () => {
  const accountService = mock<AccountService>();

  beforeEach(() => {
    mockReset(accountService);
  });

  describe("deleteAccount", () => {
    test("Успешное определение доступа на удаление аккаунта", async () => {
      const permissionService = new PermissionService({ accountService });

      accountService.determineAccountRole.mockResolvedValueOnce({
        name: "admin",
        permissions: {
          [PERMISSIONS.deleteAccount]: true,
        },
        createdAt: new Date(),
        updatedAt: new Date(),
        isDeleted: false,
      });

      const isAvailableToDeleteAccount = await permissionService.deleteAccount(DEFAULT_ACCOUNT);

      expect(accountService.determineAccountRole).toHaveBeenCalledWith(DEFAULT_ACCOUNT);
      expect(accountService.determineAccountRole).toHaveBeenCalledTimes(1);
      expect(isAvailableToDeleteAccount).toBe(true);
    });

    test("У пользователя нет прав на удаление", async () => {
      const permissionService = new PermissionService({ accountService });

      accountService.determineAccountRole.mockResolvedValueOnce({
        name: "admin",
        permissions: {
          read: true,
        },
        createdAt: new Date(),
        updatedAt: new Date(),
        isDeleted: false,
      });

      const isAvailableToDeleteAccount = await permissionService.deleteAccount(DEFAULT_ACCOUNT);

      expect(accountService.determineAccountRole).toHaveBeenCalledWith(DEFAULT_ACCOUNT);
      expect(accountService.determineAccountRole).toHaveBeenCalledTimes(1);
      expect(isAvailableToDeleteAccount).toBe(false);
    });

    test("Ошибка должна успешно пробрасываться", async () => {
      const permissionService = new PermissionService({ accountService });

      accountService.determineAccountRole.mockRejectedValueOnce(new Error("role not found"));

      expect(permissionService.deleteAccount(DEFAULT_ACCOUNT)).resolves.toBe(false);
      expect(accountService.determineAccountRole).toHaveBeenCalledWith(DEFAULT_ACCOUNT);
      expect(accountService.determineAccountRole).toHaveBeenCalledTimes(1);
    });
  });
});
